package ch.vorburger.jpapattern.framework;

import java.util.BitSet;

import javax.persistence.EntityManager;

import org.apache.openjpa.enhance.PersistenceCapable;
import org.apache.openjpa.kernel.Broker;
import org.apache.openjpa.kernel.OpenJPAStateManager;
import org.apache.openjpa.meta.FieldMetaData;
import org.apache.openjpa.util.OpenJPAId;

/**
 * JPA helper.
 * 
 * Exposes two functions not available via the standard JPA API. 
 * 
 * @author Michael Vorburger (MVO)
 */
public abstract class JPAHelper {
	private JPAHelper() {
	}
	
	/**
	 * Helper to check if a given field is loaded in a given entity instance.
	 * 
	 * Fields in JPA are loaded e.g. if they are annotated as FetchType.EAGER,
	 * or if a JOIN FETCH in a Query requested them to be, or if some JPA
	 * implementation specific API (such as an OpenJPA FetchPlan) asked for them
	 * to be.
	 * 
	 * @param entity JPA Entity instance
	 * @param fieldName name of a field in the Entity
	 * @return true if the field is loaded in the entity, false if not
	 */
	public abstract boolean isFieldLoaded(Object entity, String fieldName);

	/**
	 * Helper to obtain a JPA "reference" object for an association field.
	 * 
	 * The idea is that an ORM typically has the foreign key (FK) of an
	 * association/reference field loaded, even if it keeps the respective
	 * field null.
	 * 
	 * @see EntityManager#getReference(Class, Object)
	 *  
	 * @param entity a loaded JPA Entity instance
	 * @param fieldName name of a field of type association to another Entity 
	 * @param entityClass class of the associated field Entity
	 * @return Reference, or null if the fieldName refers to a null field
	 */
	public abstract <T> T getReference(Object entity, String fieldName, Class<T> entityClass);
	
	

	public static final JPAHelper OPENJPA = new OpenJPAHelper();

	/**
	 * An implementation of the JPAHelper. This one is highly-specific to
	 * OpenJPA and uses some internal APIs - but it should be possible to
	 * implement something similar for other JPA implementations such as
	 * EclipseLink or Hibernate.
	 */
	private static class OpenJPAHelper extends JPAHelper {
		
		@Override
		public boolean isFieldLoaded(Object entity, String fieldName) {
	        OpenJPAStateManager stateManager = getSM(entity);
	        int index = getFieldMetaDataIndex(entity, fieldName, stateManager);
	        BitSet loaded = stateManager.getLoaded();
	        return loaded.get(index);
		}
		
		@Override
	    public <T> T getReference(Object entity, String fieldName, Class<T> entityClass) {
	        Object id = getAssociationId(entity, fieldName);
	        T ref = getReference(entityClass, id, entity);
	        return ref;
	    }

		
	    private static Object getAssociationId(Object entity, String fieldName) {
	        OpenJPAStateManager stateManager = getSM(entity);
	        int index = getFieldMetaDataIndex(entity, fieldName, stateManager);
	        Object objId = stateManager.getIntermediate(index);
	        // objId CAN be null if the field is optional and the DB has a NULL in the column
	        if (objId != null) {
	            return ((OpenJPAId) objId).getIdObject();
	        } else {
	            return null;
	        }
	    }
	    
	    private static int getFieldMetaDataIndex(Object entity, String fieldName, OpenJPAStateManager stateManager) {
	        FieldMetaData fieldMetaData = stateManager.getMetaData().getDeclaredField(fieldName);
	        if (fieldMetaData != null) {
	        	return fieldMetaData.getIndex();
	        } else {
	        	throw new JPAHelperException("Field \"" + fieldName + "\" is not found in the metadata of the " + entity.getClass() + " entity?!");
	        }
		}
	    
		private static OpenJPAStateManager getSM(Object entity) {
	        PersistenceCapable pc = PersistenceCapable.class.cast(entity); // fails if not enhanced
	        OpenJPAStateManager stateManager = (OpenJPAStateManager) pc.pcGetStateManager();
	        if (stateManager == null || stateManager.isDetached()) {
	        	throw new JPAHelperException("StateManager is null or already detached, for the entity: " + entity.toString());
	        }
	        return stateManager;
	    }
	    
		/**
		 * Like getReference of EntityManager,
		 * but works given only an entity, without EntityManager.
		 * 
		 * Needed because the idea is to use getReference (above) from within a @PostLoad,
		 * where it's not normally / directly possible to get access to current EntityManager
		 * (unless you keep it around in some ThreadLocal or something). 
		 * 
		 * @see EntityManager#getReference(Class, Object)
		 * 
		 * @param entityClass class of Entity
		 * @param id JPA Id (Basic or Compound)
		 * @param entity an attached persistence capable OpenJPA entity
		 * @return Entity Reference
		 */
	    private static <T> T getReference(Class<T> entityClass, Object id, Object entity) {
	        OpenJPAStateManager stateManager = getSM(entity);
	        Broker broker = (Broker) stateManager.getContext();
	        // @see EntityManagerImpl#getReference
	        Object oid = broker.newObjectId(entityClass, id);
	        return entityClass.cast(broker.find(oid, false, null));
	    }
	}
}
