package ch.vorburger.jpapattern.framework.aop;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;
import ch.vorburger.jpapattern.framework.NullEntity;
import ch.vorburger.jpapattern.framework.Nullable;

/**
 * Factory to create NullEntity objects.
 * 
 * @see NullEntity
 * @see Nullable
 * 
 * @author Yann Andenmatten
 */
public final class NullEntityFactory {

    private static Map<Class<?>, Nullable> nullObjectSingletons = new HashMap<Class<?>, Nullable>();

    private NullEntityFactory() {
    }
    
    /**
     * Given an Entity class, this returns a "NullEntity". 
     *
	 * The object returned for a given clazz
	 * respects the "contract" described in NullEntity & Nullable,
	 * and is always the same (per class), it's a Singleton.
	 * 
     * @param clazz Entity class, must implement Nullable
     * @return an object which is a subclass of the clazz and which implements NulLEntity 
     */
    @SuppressWarnings("unchecked")
	public static <T extends Nullable> T theNullObject(Class<T> clazz) {
    	if (!nullObjectSingletons.containsKey(clazz)) {
    		Set<Class<?>> interfaces = asSet(clazz.getInterfaces());
    		interfaces.add(NullEntity.class); // (yan) Still relevant to have a interface for that ?
    		nullObjectSingletons.put(clazz, (NullEntity)createProxy(clazz, interfaces.toArray(new Class<?>[interfaces.size()])));
    	}
    	return (T)nullObjectSingletons.get(clazz);
    }
    
	@SuppressWarnings("unchecked")
	private static <T> T createProxy(Class<T> clazz, Class<?>[] interfaces) {
		Enhancer enhancer = new Enhancer();
		enhancer.setSuperclass(clazz);
		enhancer.setInterfaces(interfaces);
		enhancer.setCallback(new MethodInterceptor() {
			@Override
			//CHECKSTYLE:OFF throws Throwable is OK here, as cglib's invokeSuper throws Throwable
			public Object intercept(Object obj, Method method, Object[] args, MethodProxy proxy) throws Throwable {
			//CHECKSTYLE:ON				
				if (method.getName().equals("isNull")) {
					return Boolean.TRUE;
				} else if (method.getName().startsWith("set")) {
					// TODO Should throw NullEntity.UOE_MSG, with args, like NullCEntity example
					throw new UnsupportedOperationException("setter forbidden");
				} else if (method.getName().startsWith("get")) {
					if (asSet(method.getReturnType().getInterfaces()).contains(Nullable.class)) {
						Class<? extends Nullable> nullObjectType = (Class<? extends Nullable>)method.getReturnType();
						return theNullObject(nullObjectType); // uses default constructor
					} else {
						return proxy.invokeSuper(obj, args);
					}
				} else {
					return proxy.invokeSuper(obj, args);
				}
			}
			
		});
		return (T) enhancer.create();
	}
	
	private static Set<Class<?>> asSet(Class<?>[] interfaces) {
		Set<Class<?>> interfacesSet = new HashSet<Class<?>>();
		interfacesSet.addAll(Arrays.asList(interfaces));
		return interfacesSet;
	}
    
}
