package ch.vorburger.jpapattern.sample;

import java.io.File;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;

import junit.framework.TestCase;

import org.apache.commons.io.FileUtils;
import org.apache.openjpa.lib.jdbc.AbstractJDBCListener;
import org.apache.openjpa.lib.jdbc.JDBCEvent;
import org.apache.openjpa.lib.jdbc.JDBCListener;
import org.apache.openjpa.persistence.OpenJPAEntityManager;
import org.apache.openjpa.persistence.OpenJPAPersistence;

import ch.vorburger.jpapattern.framework.NullEntity;
import ch.vorburger.jpapattern.sample.entities.AEntity;

/**
 * Test Case & Demo/Tutorial.
 *  
 * @author Michael Vorburger
 */
public class JPAPatternExampleTest extends TestCase {
    
    private static final String DB = "target/db";
    private List<String> sql = new LinkedList<String>();
    
    @SuppressWarnings("unchecked")
    public void testPattern() throws Exception  {
        EntityManagerFactory emf = createEMF();
        createTables(emf);
        try {
        	sql.clear();
            EntityManager em = emf.createEntityManager();
            List<AEntity> list = em.createQuery("SELECT o FROM AEntity o ORDER BY o.id").getResultList();

            // Close EM now, to a) make sure entities in list are detached and
            // no lazy loading here and b) to make sure this pattern actually works
            // for real-world detached use cases
            em.close();
            assertEquals(2, list.size());

            // Read checkIds method to understand what you can do:
			checkIds(list);
			
            // You can NOT (yet) do this of course, as the type field was lazy!
			try {
				list.get(0).getB().getName();
				fail();
			} catch (IllegalStateException ex) {
				// As expected!
			}
			
			// You can not set fields of Reference objects, once detached (in OpenJPA at least).
			// list.get(0).getAnotherType().setCode("XYZ");
			
			checkSetter(list.get(1));
			
            // To wrap-up, let's make sure all SQL is as expected
            assertEquals(1, sql.size());
            assertEquals("SELECT t0.id, t0.c_id FROM test_complex t0 ORDER BY t0.id ASC", sql.get(0));
            
			
            // Now let's use Fetch to actually get related entities (SQL JOIN)
            sql.clear();
            em = emf.createEntityManager();
            list = em.createQuery("SELECT o FROM AEntity o LEFT OUTER JOIN FETCH o.b ORDER BY o.id").getResultList();

			/*
			 * Instead of the JPA standard JOIN FETCH, the OpenJPA specific
			 * FetchGroup API could have been used as well: 
			 * OpenJPAEntityManager oem = (OpenJPAEntityManager) em;
			 * oem.getFetchPlan().addField(AEntity.class, "b");
			 * list = em.createQuery("SELECT o FROM AEntity o ORDER BY o.id").getResultList(); 
			 * 
			 * One of the advantages of the API approach is that it keeps the "filter"
			 * separate from the "fetch plan". It's also seems a bit stupid how
			 * the JPQL FETCH requires an explicit OUTER, whereas the API
			 * approach derives implicitly this from optional / nullable
			 */            
			em.close(); // as above
            assertEquals(2, list.size());

            // First make sure Ids still work
            checkIds(list);

            // Now see that thanks to the FetchGroup we can actually access Type.code fields:
            AEntity firstResult = list.get(0);
			AEntity secondResult = list.get(1);

            assertEquals("GHI", firstResult.getB().getName());

            
            // And lastly, how to explicitly check (if you really have to know), either like this...
            assertFalse(firstResult.getB().isNull());
            assertTrue(secondResult.getB().isNull());
            // ... or like this...
            assertFalse(firstResult.getB() instanceof NullEntity);
            assertTrue(secondResult.getB() instanceof NullEntity);

			// And finally, test the transitive and AOP-based null'd D/DSub
            assertTrue(secondResult.getB().getC().isNull()); // Automatic transitive null object, thanks to AOP :-)
            assertTrue((secondResult.getB().getC()) instanceof NullEntity); // Null object type

            // To wrap-up, let's make sure all SQL is as expected
            assertEquals(1, sql.size());
            assertEquals("SELECT t0.id, t1.id, t1.d_id, t1.name FROM test_complex t0 LEFT OUTER JOIN test_c t1 ON t0.c_id = t1.id ORDER BY t0.id ASC", sql.get(0));

        } finally {
            emf.close();
        }
    }

	private void checkSetter(AEntity entity) {
		// You can, obviously, NOT set fields of NULL objects, as there is no row in the RDBMS for this
		try {
			entity.getB().setName("???");
			fail();
		} catch (UnsupportedOperationException ex) {
			// As expected
		}
	}

	private void checkIds(List<AEntity> list) {
		// JPA Entities could not be used in this manner without applying this pattern: 
		
		AEntity firstResult = list.get(0);
		assertEquals(Long.valueOf(321), firstResult.getB().getId()); // yupee!

		AEntity secondResult = list.get(1);
		assertNull(secondResult.getB().getId()); // null object pattern rocks!
	}


    private void createTables(EntityManagerFactory emf) throws Exception {
        EntityManager em = emf.createEntityManager();
        em.getTransaction().begin();
        OpenJPAEntityManager kem = OpenJPAPersistence.cast(em);
        Connection conn = (Connection) kem.getConnection();

        executeStatement(conn, "INSERT INTO test_c(id, name) VALUES (321, 'GHI')");

        executeStatement(conn, "INSERT INTO test_complex(id, c_id) VALUES (1, 321)");
        executeStatement(conn, "INSERT INTO test_complex(id, c_id) VALUES (2, NULL)");

        conn.close();
        
        em.flush();
        em.getTransaction().commit();
        em.close();
    }

    private EntityManagerFactory createEMF() throws IOException {
        FileUtils.deleteDirectory(new File(DB));
        
        Properties properties = new Properties();
        properties.setProperty("openjpa.ConnectionDriverName", "org.apache.derby.jdbc.EmbeddedDriver");
        properties.setProperty("openjpa.ConnectionURL", "jdbc:derby:" + DB + ";create=true");

        properties.setProperty("openjpa.RuntimeUnenhancedClasses", "unsupported");
        properties.setProperty("openjpa.DetachState", "loaded(AccessUnloaded=false)");
        
        properties.setProperty("openjpa.Log", "DefaultLevel=WARN, Runtime=INFO, Tool=INFO, SQL=TRACE");
        properties.setProperty("openjpa.ConnectionFactoryProperties", "PrettyPrint=true, PrettyPrintLineLength=100");
        
        properties.setProperty("openjpa.jdbc.SynchronizeMappings", "buildSchema(ForeignKeys=true)");

        
        properties.put("openjpa.jdbc.JDBCListeners", new JDBCListener[] { this.new Listener() });

        
        EntityManagerFactory emf = Persistence.createEntityManagerFactory("test", properties);
        return emf;
    }

    private void executeStatement(Connection connection, String sqlStatement) throws SQLException {
        Statement statement = connection.createStatement();
        statement.execute(sqlStatement);
		statement.close();
	}

	public class Listener extends AbstractJDBCListener {
        @Override
        public void beforeExecuteStatement(JDBCEvent event) {
            if (event.getSQL() != null && sql != null) {
                sql.add(event.getSQL());
            }
        }
    }
}
